package zero1hd.rhythmbullet.desktop.graphics.ui.components;

import static org.lwjgl.openal.AL10.*;

import java.nio.ByteBuffer;
import java.nio.ShortBuffer;

import org.lwjgl.openal.AL11;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.backends.lwjgl.audio.OpenALMusic;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.scenes.scene2d.ui.Widget;
import com.badlogic.gdx.utils.Disposable;
import com.badlogic.gdx.utils.reflect.ClassReflection;
import com.badlogic.gdx.utils.reflect.Field;
import com.badlogic.gdx.utils.reflect.ReflectionException;

import zero1hd.rhythmbullet.audio.visualizer.BasicVisualizer;
import zero1hd.rhythmbullet.util.MusicManager;

public class Visualizer extends Widget implements Disposable {
	private BasicVisualizer vis;
	private boolean updatePositioning = true;
	private boolean mmSet;
	private MusicManager mm;
	private ShortBuffer playingBuffer;
	private ShortBuffer compareBuffer;
	private ShortBuffer buffer;
	private int sourceID;
	private float visRefreshRate;
	private float timer;
	private int readWindowIndex;
	public Visualizer() {
		vis = new BasicVisualizer();
		
		try {
			Field bufferField = ClassReflection.getDeclaredField(OpenALMusic.class, "tempBuffer");
			bufferField.setAccessible(true);
			buffer = ((ByteBuffer) bufferField.get(null)).asShortBuffer().asReadOnlyBuffer();
		} catch (IllegalArgumentException | SecurityException | ReflectionException e) {
			e.printStackTrace();
			Gdx.app.debug("Visualizer reflection", "Failed attempt at retrieving tempBuffer field.");
			Gdx.app.exit();
		}
		
	}
	
	@Override
	public void draw(Batch batch, float parentAlpha) {
		vis.render(batch, parentAlpha);
		super.draw(batch, parentAlpha);
	}
	
	@Override
	public void act(float delta) {
		if (mm != null && mm.isFinishedLoading() && !mmSet) {
			vis.setMM(mm);
			visRefreshRate = mm.getReadWindowSize()/mm.getSampleRate();
			mmSet = true;
			Gdx.app.debug("Visualizer", "\nsample count: " + mm.getSampleCount()
			+ "\nDuration in seconds: " + mm.getDuration() + 
			"\nSample rate: " + mm.getSampleRate() +
			"\nRefresh rate: " + visRefreshRate);
		}
		
		vis.update(delta);
		updateVisualizerProperties();
		if (updatePositioning) {
			vis.updatePositionInfo();
			vis.setxPos((getWidth() - vis.getActualWidth())/2f);
		}
		if (mmSet) {
			if (timer >= visRefreshRate) {
				timer = 0;
				calcPCMData();
				vis.calculate(delta);
			} else {
				timer += delta;
			}
		}
		super.act(delta);
	}
	
	public void calcPCMData() {
		short chanVal;
		int bufferPosOffset = 0;
		if (mm.playbackIndexUpdate() != readWindowIndex) {
			bufferPosOffset = updateBufferPosition();
			if (bufferPosOffset < 0 && playingBuffer.limit()+bufferPosOffset > 0) {
				playingBuffer.position(playingBuffer.limit()+bufferPosOffset);
			}
		}
		
		for (int sid = 0; sid < vis.getAudioPCM().length && sid < buffer.remaining(); sid++) {
			for (int channel = 0; channel < mm.getChannelCount(); channel ++) {
				if (playingBuffer.position() < playingBuffer.limit()) {
					if (vis.getAudioPCM()[sid] < (chanVal = playingBuffer.get())) {
						vis.getAudioPCM()[sid] = chanVal;
					}
				} else {
					if (vis.getAudioPCM()[sid] < (chanVal = buffer.get())) {
						vis.getAudioPCM()[sid] = chanVal;
					}
				}
			}
			vis.getAudioPCM()[sid] /= Short.MAX_VALUE+1f;
		}
		readWindowIndex++;
		
		
		//Take down original buffer position so we don't need to sync again after...
		int originalPos = buffer.position();
		
		//Begin comparison
		buffer.rewind();
		if (compareBuffer.compareTo(buffer) != 0) {
			bufferChanged();
			
			
			//Begin copying current buffer to the comparison buffer
			compareBuffer.clear();
			compareBuffer.put(buffer);
			compareBuffer.flip();
		}
		
		//Reset buffer to proper position.
		buffer.position(originalPos);
	}
	
	private void bufferChanged() {
		playingBuffer.clear();
		playingBuffer.put(compareBuffer);
	}
	
	private int updateBufferPosition() {
		int pos = (int) ((alGetSourcef(sourceID, AL11.AL_SAMPLE_OFFSET))-buffer.limit()-mm.getChannelCount()*mm.getReadWindowSize());
		readWindowIndex = mm.getPlaybackIndexPosition()-1;
		if (pos < 0) {
			buffer.position(0);
		} else {
			buffer.position(pos);
		}
		return pos;
	}
	
	public void setMM(MusicManager mm) {
		try {
			Field sourceIDField = ClassReflection.getDeclaredField(OpenALMusic.class, "sourceID");
			sourceIDField.setAccessible(true);
			sourceID = (int) sourceIDField.get(mm.getMusic());
		} catch (ReflectionException e) {
			e.printStackTrace();
		}
		int originalPos = buffer.position();

		playingBuffer = ShortBuffer.allocate(buffer.capacity());
		buffer.rewind();
		playingBuffer.put(buffer);
		playingBuffer.flip();
		
		compareBuffer = ShortBuffer.allocate(buffer.capacity());
		buffer.rewind();
		compareBuffer.put(buffer);
		compareBuffer.flip();
		
		buffer.position(originalPos);
		this.mm = mm;
		mmSet = false;
	}
	
	@Override
	public void setColor(Color color) {
		vis.setColor(color);
		super.setColor(color);
	}
	
	@Override
	public void setColor(float r, float g, float b, float a) {
		vis.setColor(r, g, b, a);
		super.setColor(r, g, b, a);
	}
	
	public void updateVisualizerProperties() {
		vis.setHeight(getHeight());
		vis.setWidth(getWidth());
		vis.setxPos(getX());
		vis.setyPos(getY());
		vis.setRotation(getRotation());
	}
	
	public void setUpdatePositioning(boolean updatePositioning) {
		updateVisualPosition();
		this.updatePositioning = updatePositioning;
	}
	
	public void updateVisualPosition() {
		updateVisualizerProperties();
		vis.updatePositionInfo();
		vis.setxPos(((vis.getWidth() - vis.getActualWidth())/2f));
		vis.updatePositionInfo();
	}
	
	public boolean isUpdatePositioning() {
		return updatePositioning;
	}
	
	public BasicVisualizer getVis() {
		return vis;
	}

	@Override
	public void dispose() {
		vis.dispose();
	}
	
}
